from __future__ import annotations

import collections
import datetime
import decimal
import typing
import uuid

from dataclasses_avroschema import types

from . import fields, pydantic_fields

INMUTABLE_FIELDS_CLASSES: dict[type | str, type] = {
    bool: fields.BooleanField,
    int: fields.LongField,
    types.Int32: fields.IntField,
    float: fields.DoubleField,
    types.Float32: fields.FloatField,
    bytes: fields.BytesField,
    str: fields.StringField,
    type(None): fields.NoneField,
}

CONTAINER_FIELDS_CLASSES = {
    tuple: fields.TupleField,
    list: fields.ListField,
    collections.abc.Sequence: fields.ListField,
    collections.abc.MutableSequence: fields.ListField,
    dict: fields.DictField,
    collections.abc.Mapping: fields.DictField,
    collections.abc.MutableMapping: fields.DictField,
    typing.Union: fields.UnionField,
}

LOGICAL_TYPES_FIELDS_CLASSES = {
    datetime.date: fields.DateField,
    datetime.time: fields.TimeMilliField,
    types.TimeMicro: fields.TimeMicroField,
    datetime.datetime: fields.DatetimeField,
    types.DateTimeMicro: fields.DatetimeMicroField,
    uuid.uuid4: fields.UUIDField,
    uuid.UUID: fields.UUIDField,
    bytes: fields.BytesField,
}

SPECIAL_ANNOTATED_TYPES = {
    decimal.Decimal: fields.DecimalField,
    types.Fixed: fields.FixedField,
}

PYDANTIC_TYPES = None

try:
    import pydantic  # pragma: no cover

    PYDANTIC_INMUTABLE_FIELDS_CLASSES = {
        pydantic.FilePath: pydantic_fields.FilePathField,
        pydantic.DirectoryPath: pydantic_fields.DirectoryPathField,
        pydantic.EmailStr: pydantic_fields.EmailStrField,
        pydantic.NameEmail: pydantic_fields.NameEmailField,
        pydantic.AnyUrl: pydantic_fields.AnyUrlField,
        pydantic.AnyHttpUrl: pydantic_fields.AnyHttpUrlField,
        pydantic.HttpUrl: pydantic_fields.HttpUrlField,
        pydantic.FileUrl: pydantic_fields.FileUrlField,
        pydantic.PostgresDsn: pydantic_fields.PostgresDsnField,
        pydantic.CockroachDsn: pydantic_fields.CockroachDsnField,
        pydantic.AmqpDsn: pydantic_fields.AmqpDsnField,
        pydantic.RedisDsn: pydantic_fields.RedisDsnField,
        pydantic.MongoDsn: pydantic_fields.MongoDsnField,
        pydantic.KafkaDsn: pydantic_fields.KafkaDsnField,
        pydantic.SecretStr: pydantic_fields.SecretStrField,
        pydantic.IPvAnyAddress: pydantic_fields.IPvAnyAddressField,
        pydantic.IPvAnyInterface: pydantic_fields.IPvAnyInterfaceField,
        pydantic.IPvAnyNetwork: pydantic_fields.IPvAnyNetworkField,
        pydantic.NegativeFloat: pydantic_fields.NegativeFloatField,
        pydantic.PositiveFloat: pydantic_fields.PositiveFloatField,
        pydantic.NegativeInt: pydantic_fields.NegativeIntField,
        pydantic.PositiveInt: pydantic_fields.PositiveIntField,
        pydantic.ConstrainedInt: pydantic_fields.ConstrainedIntField,
        # ConstrainedIntValue is a dynamic type that needs to be referenced by qualified name
        # and cannot be imported directly
        "ConstrainedIntValue": pydantic_fields.ConstrainedIntField,
    }

    PYDANTIC_LOGICAL_TYPES_FIELDS_CLASSES = {
        pydantic.UUID1: pydantic_fields.UUID1Field,
        pydantic.UUID3: pydantic_fields.UUID3Field,
        pydantic.UUID4: pydantic_fields.UUID4Field,
        pydantic.UUID5: pydantic_fields.UUID5Field,
    }

    INMUTABLE_FIELDS_CLASSES.update(PYDANTIC_INMUTABLE_FIELDS_CLASSES)
    LOGICAL_TYPES_FIELDS_CLASSES.update(PYDANTIC_LOGICAL_TYPES_FIELDS_CLASSES)  # type: ignore
    PYDANTIC_TYPES = (
        pydantic.FilePath,
        pydantic.DirectoryPath,
        pydantic.EmailStr,
        pydantic.NameEmail,
        pydantic.AnyUrl,
        pydantic.AnyHttpUrl,
        pydantic.HttpUrl,
        pydantic.FileUrl,
        pydantic.PostgresDsn,
        pydantic.CockroachDsn,
        pydantic.AmqpDsn,
        pydantic.RedisDsn,
        pydantic.MongoDsn,
        pydantic.KafkaDsn,
        pydantic.SecretStr,
        pydantic.IPvAnyAddress,
        pydantic.IPvAnyInterface,
        pydantic.IPvAnyNetwork,
        pydantic.NegativeFloat,
        pydantic.PositiveFloat,
        pydantic.NegativeInt,
        pydantic.PositiveInt,
        pydantic.ConstrainedInt,
        pydantic.UUID1,
        pydantic.UUID3,
        pydantic.UUID4,
        pydantic.UUID5,
    )

except ImportError:  # type: ignore # pragma: no cover
    # pydantic is not installed so we do not include their fields
    ...  # type: ignore # pragma: no cover



try:
    from faker import Faker

    fake = Faker()
except ImportError:  # pragma: no cover

    class FakeStub:
        """A stub for Faker.fake() when the feature [faker] is not enabled.

        It raises a hard RuntimeError when faker is not installed and
        we tried to invoke `.fake()` on any of our fields.

        We can find the local usages of faker methods (i.e. fake.<method>(*args, **kwargs)) with:

        ```sh
        rg "^.*fake\.(\w+)\(.*\).*$" dataclasses_avroschema -r '$1' -NIo | sort -u
        ```
        """

        def raise_runtime_error(self, *args, **kwargs):
            raise RuntimeError(
                "faker must be installed in order to use .fake(). "
                "Consider running `pip install dataclasses-avroschema[faker]`"
            )

    _FAKER_METHODS_TO_STUB = [
        "company_email",
        "date_object",
        "date_time",
        "first_name",
        "ipv4",
        "last_name",
        "pybool",
        "pydecimal",
        "pyfloat",
        "pyint",
        "pystr",
        "time_object",
        "uri",
        "url",
    ]

    fake = FakeStub()  # type: ignore

    for method in _FAKER_METHODS_TO_STUB:
        setattr(fake, method, fake.raise_runtime_error)



from typing import Any, Callable, Optional, Type, TypeVar

from fastavro.validation import validate

from dataclasses_avroschema.schema_generator import AvroModel
from dataclasses_avroschema.types import JsonDict
from dataclasses_avroschema.utils import standardize_custom_type

try:
    from pydantic import BaseModel  # pragma: no cover
except ImportError as ex:  # pragma: no cover
    raise Exception("pydantic must be installed in order to use AvroBaseModel") from ex  # pragma: no cover

CT = TypeVar("CT", bound="AvroBaseModel")


class AvroBaseModel(BaseModel, AvroModel):  # type: ignore
    @classmethod
    def generate_dataclass(cls: Type[CT]) -> Type[CT]:
        return cls

    @classmethod
    def json_schema(cls: Type[CT], *args: Any, **kwargs: Any) -> str:
        return cls.schema_json(*args, **kwargs)

    def asdict(self, standardize_factory: Optional[Callable[..., Any]] = None) -> JsonDict:
        """
        Document this. asdict vs dict
        """
        data = dict(self)

        standardize_method = standardize_factory or standardize_custom_type

        # te standardize called can be replaced if we have a custom implementation of asdict
        # for now I think is better to use the native implementation
        return standardize_method(data)

    def validate_avro(self) -> bool:
        """
        Validate that instance matches the avro schema
        """
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    @classmethod
    def fake(cls: Type[CT], **data: Any) -> CT:
        """
        Creates a fake instance of the model.

        Attributes:
            data: Dict[str, Any] represent the user values to use in the instance
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return cls.parse_obj(payload)
